/*-
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* for strdup() */
#define _POSIX_C_SOURCE 200809L

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "cache/cache.h"
#include "vcl.h"
#include "vre.h"
#include "cache/cache_director.h"

#include "patricia.h"

#define VFAIL(ctx, fmt, ...) \
	VRT_fail((ctx), "vmod selector failure: " fmt, __VA_ARGS__)

#define VERR(ctx, fmt, ...) \
	VSLb((ctx)->vsl, SLT_VCL_Error, "vmod selector error: " fmt,	\
	     __VA_ARGS__)

struct entry {
	unsigned	magic;
#define VMOD_SELECTOR_ENTRY_MAGIC 0x733dbe63
	char		*string;
	VCL_BACKEND	backend;
	vre_t		*re;
};

struct vmod_selector_set {
	unsigned	magic;
#define VMOD_SELECTOR_SET_MAGIC 0x838979ef
	struct entry	**table;
	char		**members;
	struct pt_y	*origo;
	char		*vcl_name;
	unsigned	nmembers;
	unsigned	nentries;
	VCL_BOOL	case_sensitive;
};

/* Event function */

int
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
        (void) ctx;
        (void) priv;

        if (e == VCL_EVENT_LOAD && !PT_Inited())
		PT_Init();
        return 0;
}

/* Object regex */

VCL_VOID
vmod_set__init(VRT_CTX, struct vmod_selector_set **setp, const char *vcl_name,
	       VCL_BOOL case_sensitive)
{
	struct vmod_selector_set *set;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(setp);
	AZ(*setp);
	AN(vcl_name);
	
	ALLOC_OBJ(set, VMOD_SELECTOR_SET_MAGIC);
	AN(set);
	*setp = set;
	set->vcl_name = strdup(vcl_name);
	AN(set->vcl_name);
	set->case_sensitive = case_sensitive;
	AZ(set->table);
	AZ(set->members);
	AZ(set->origo);
	AZ(set->nentries);
}

VCL_VOID
vmod_set__fini(struct vmod_selector_set **setp)
{
	struct vmod_selector_set *set;

	if (setp == NULL || *setp == NULL)
		return;
	CHECK_OBJ(*setp, VMOD_SELECTOR_SET_MAGIC);
	set = *setp;
	*setp = NULL;
	PT_Free(set->origo);
	for (unsigned i = 0; i < set->nmembers; i++)
		free(set->members[i]);
	for (unsigned i = 0; i < set->nentries; i++) {
		struct entry *entry = set->table[i];
		CHECK_OBJ_NOTNULL(entry, VMOD_SELECTOR_ENTRY_MAGIC);
		free(entry->string);
		if (entry->re != NULL)
			VRE_free(&entry->re);
		FREE_OBJ(entry);
	}
	free(set->members);
	free(set->table);
	free(set->vcl_name);
	FREE_OBJ(set);
}

VCL_VOID
vmod_set_add(VRT_CTX, struct vmod_selector_set *set, VCL_STRING member,
	     VCL_STRING string, VCL_STRING regex, VCL_BACKEND backend)
{
	struct entry *entry;
	unsigned n;
	vre_t *re = NULL;
	const char *error;
	int erroffset;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);
	CHECK_OBJ_ORNULL(backend, DIRECTOR_MAGIC);

	if ((ctx->method & VCL_MET_INIT) == 0) {
		VFAIL(ctx, "%s.add() may only be called in vcl_init",
		      set->vcl_name);
		return;
	}

	if (member == NULL) {
		VFAIL(ctx, "%s.add(): string to be added is NULL",
		      set->vcl_name);
		return;
	}

	set->nmembers++;
	n = set->nmembers;
	set->members = realloc(set->members, n * sizeof(VCL_STRING));
	AN(set->members);
	set->members[n - 1] = strdup(member);
	AN(set->members[n - 1]);

	errno = 0;
	if (PT_Insert(&set->origo, n - 1, set->members) != 0) {
		if (errno == EINVAL)
			VFAIL(ctx, "%s.add(): \"%s\" added more than once",
			      set->vcl_name, member);
		else
			VFAIL(ctx, "%s.add(\"%s\") failed: %s", set->vcl_name,
			      member, strerror(errno));
		return;
	}

	if (regex != NULL) {
		/* XXX expose VRE options */
		re = VRE_compile(regex, 0, &error, &erroffset);
		if (re == NULL) {
			VFAIL(ctx, "%s.add(): cannot compile regular expression"
			      " '%s': %s at offset %d", set->vcl_name, regex,
			      error, erroffset);
			return;
		}
	}

	if (string == NULL && re == NULL && backend == NULL)
		return;

	set->nentries = n;
	set->table = realloc(set->table, n * sizeof(struct entry *));
	AN(set->table);
	ALLOC_OBJ(entry, VMOD_SELECTOR_ENTRY_MAGIC);
	AN(entry);
	if (string != NULL)
		entry->string = strdup(string);
	if (re != NULL)
		entry->re = re;
	if (backend != NULL)
		entry->backend = backend;
	set->table[n - 1] = entry;
}

VCL_BOOL
vmod_set_match(VRT_CTX, struct vmod_selector_set *set, VCL_STRING subject)
{
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	if (subject == NULL) {
		VERR(ctx, "%s.match(): subject string is NULL", set->vcl_name);
		return (0);
	}
	if (set->nmembers == 0) {
		VERR(ctx, "%s.match(): no entries were added", set->vcl_name);
		return (0);
	}

	AN(set->origo);
	AN(set->members);
	return (PT_Lookup(set->origo, set->members, subject) != UINT_MAX);
}

VCL_STRING
vmod_set_debug(VRT_CTX, struct vmod_selector_set *set)
{
	struct vsb *sb;
	VCL_STRING output;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	sb = PT_Dump(set->origo, set->members);
	CHECK_OBJ_NOTNULL(sb, VSB_MAGIC);
	output = WS_Copy(ctx->ws, VSB_data(sb), -1);
	VSB_destroy(&sb);
	if (output == NULL)
		VFAIL(ctx, "%s.debug(): out of workspace", set->vcl_name);
	return output;
}

VCL_STRING
vmod_version(VRT_CTX)
{
	(void) ctx;
	return VERSION;
}
